/// <reference path="typings/tp.d.ts">/>
/// <reference path="typings/game-object.d.ts">/>

/** START BTN JSON **/
const BUTTON_OBJ_JSON = `{
  "objectType": "DynamicObject",
  "transform": {
    "rotation": {
      "x": 0.0,
      "y": 0.0,
      "z": 0.0,
      "w": 0.0
    },
    "translation": {
      "x": 0.0,
      "y": 0.0,
      "z": 0.0
    },
    "scale3D": {
      "x": 1,
      "y": 1,
      "z": 1
    }
  },
  "simulatingPhysics": true,
  "primaryColor": {
    "b": 0,
    "g": 0,
    "r": 0,
    "a": 255
  },
  "secondaryColor": {
    "b": 0,
    "g": 0,
    "r": 0,
    "a": 255
  },
  "metallic": 0,
  "roughness": 0,
  "friction": 0,
  "restitution": 0,
  "density": 1,
  "surfaceType": "SurfaceType1",
  "objectName": "",
  "objectDescription": "",
  "collisionType": "CB_Regular",
  "templateId": "B64FD768425FADCF829D8FBB811751C9",
  "shouldSnap": true,
  "previousPosition": {
    "x": 0,
    "y": 0,
    "z": 0
  },
  "objectScriptPackage": "49CF927A4DEADB40D012BAA7F275C648",
  "objectScriptName": "",
  "uniqueId": "kho",
  "bCanBeDamaged": false
}`;
/** END BTN JSON **/

/** START SHARED **/
const SlotsColorName = ['Blue', 'Green', 'Red', 'DarkGreen', 'Purple', 'Pink', 'Orange', 'Brown', 'White', 'Yellow'];
const BasicColors = {
  white: new Color(255, 255, 255, 255),
  black: new Color(0, 0, 0, 255),
  red: new Color(255, 0, 0, 255),
  green: new Color(0, 255, 0, 255),
  blue: new Color(0, 0, 255, 255),
  yellow: new Color(255, 255, 0, 255),
  orange: new Color(255, 165, 0, 255),
  magenta: new Color(255, 0, 255, 255),
};
/** END SHARED **/

const MAX_SLOTS = 10;

/**
 * @type {GameObject[]} array of create buttons
 */
let buttonObjs = [];

/**
 * @type {GameObject} reference to attached object
 */
let refObject = this.refObject;

refObject.setName('Grab this to activate');

refObject.onCreated.add(() => createButtons());
refObject.onGrab.add(() => removeButtons())
refObject.onReleased.add(() => createButtons());

globalEvents.onPlayerJoined.add(() => refreshButtons());
globalEvents.onPlayerLeft.add(() => refreshButtons());
globalEvents.onPlayerSwitchedSlots.add(() => refreshButtons());

function createButtons() {
  const players = world.getAllPlayers();
  for (let i = 0; i < players.length; i++) {
    const player = players[i];
    const pos = calcButtonPos(i);
    createButton(pos, i, player.getPlayerColor(), swapPlayers);
  }
}

/**
 * Calc button position
 * @param {number} idx button index
 */
function calcButtonPos(idx) {
  const sc = refObject.getScale();
  const ro = refObject.getRotation();
  const po = refObject.getPosition();
  const pos = new Vector(
    sc.x * 6 * Math.sin(toRad(ro.yaw + 36 * Math.abs(10 * (idx % 2) - Math.ceil(idx / 2)))) + po.x,
    sc.y * 6 * Math.cos(toRad(ro.yaw + 36 * Math.abs(10 * (idx % 2) - Math.ceil(idx / 2)))) + po.y,
    po.z,
  ); // From TTS mod
  return pos;
}

function toRad(degrees) {
  var pi = Math.PI;
  return degrees * (pi / 180);
}

/**
 * Create new button object
 * @param {Vector} pos position of button
 * @param {number} swapSlotId index of slot for swap
 * @param {Color} color button color
 * @param {Function} callback call on release button obj
 */
function createButton(pos, swapSlotId, color, callback) {
  const obj = world.createObjectFromJSON(BUTTON_OBJ_JSON, new Vector(0, 0, 0));
  obj.setPosition(pos);
  obj.setPrimaryColor(color);
  obj.setBounciness(0);
  obj.onReleased.add((_, p) => callback(p, swapSlotId));
  buttonObjs.push(obj);
}

/**
 * Swap player between slots
 * @param {Player} clickPlayer player who click button
 * @param {number} slotId slot index to swap
 */
function swapPlayers(clickPlayer, slotId) {
  if (clickPlayer.getSlot() === slotId) {
    return;
  }

  const playerToSwap = world.getPlayerBySlot(slotId);
  const tempSlot = getFirstEmptySlot();

  if (tempSlot === null) {
    world.broadcastChatMessage('Swapper can work only if exist at least one free slot', BasicColors.red);
    return;
  }

  if (!playerToSwap) {
    clickPlayer.switchSlot(slotId);
  } else {
    clickPlayer.switchSlot(tempSlot);
    playerToSwap.switchSlot(clickPlayer.getSlot());
    clickPlayer.switchSlot(playerToSwap.getSlot());

    world.broadcastChatMessage(`Player ${clickPlayer.getName()} (${SlotsColorName[clickPlayer.getSlot()]}) swap with ${playerToSwap.getName()} (${SlotsColorName[playerToSwap.getSlot()]})`);
  }
}

function getFirstEmptySlot() {
  const busySlots = world.getAllPlayers().map(p => p.getSlot());
  for (let i = 0; i < MAX_SLOTS; i++) {
    if (!busySlots.includes(i)) {
      return i;
    }
  }
  return null;
}

function removeButtons() {
  for (const obj of buttonObjs) {
    obj.destroy();
  }
}

function refreshButtons() {
  removeButtons();
  createButtons();
}